#!/usr/bin/perl
#blastn2qrnadepth.pl

use strict;

use constant GNUPLOT => '/usr/bin/gnuplot';

if (!@ARGV) { 
        print "usage: blastn2qrnadepth.pl <file> min_len min_id max_id max_evalue depth which[ID|SC] verbose\n";
        print "min_len:   minimum length   of blast hits allowed  \n";
        print "min_id:    minimum identity of blast hits allowed \n";
        print "max_id:    maximum identity of blast hits allowed \n";
        print "max_eval:  maximum evalue   of blast hits allowed  \n";
        print "depth:     max number of alignment coverage per position\n";
        print "which:     criteria to pick alignments\n";
	print "                    ID -- best \% identity\n";
	print "                    SC -- best score\n";
        exit;
}
my $filename  = shift;
my $dir;
my $file;

if ($filename =~ /^(\S+)\/([^\/]+)$/) {
    $dir  = $1."/";
    $file = $2;
}
else {
    $dir  = "";
    $file = $filename;
}


my $min_len  = shift;
my $min_id   = shift;
my $max_id   = shift;
my $max_eval = shift;

my $depth = shift;
my $which = shift;

my $verbose;
undef $verbose;
$verbose = shift;

my $shift = 50;

my $filea  = "$file.a";    # here I dump the results of using len, evalue and identity cutoffs
my $fileb  = "$file.b";    # here I order the aligments for any query according to %id or score
my $filec  = "$file.c";    # final list of alignments in pseudoblastn format
my $fileq  = "$file.q";    # qrna-ready file
my $report = "$file.rep";

my $max_len = 3000000;


my %len_query;

my $n_query = 0;
my $n_ali_total = 0;
my $n_ali_one = 0;
my $n_ali_two = 0;

my %n_ali_per_query_total;
my %n_ali_per_query_one;
my %n_ali_per_query_two;

my $ave_len_ali_total = 0;
my $ave_len_ali_one   = 0;
my $ave_len_ali_two   = 0;
    


prune_blastn  ("$dir$file", "$filea");

if ($n_ali_one > 0) {
    order_blastn  ("$filea",     "$fileb");  system ("rm $filea\n");
    depile_blastn ("$fileb",     "$filec");  system ("rm $fileb\n");
    blastn_2_qrna ("$filec",     "$fileq");  system ("rm $filec\n");
}
else { 
    system ("rm $filea\n"); 
}

write_report  ("$report");



##
## SUBROUTINES
##

# blastn_2_qrna ()
#
#
sub blastn_2_qrna {

    my ($file, $fileq) = @_;

    my $nseq1 = 0;
    my $nseq2 = 0;
    
    my $line1 = 0;
    my $line2 = 0;
    
    my $name1;
    my $name2;

    my $annotation;

    my $more;
    my $info;

    my $start1;
    my $start2;

    my @more_parts;

    my $seq = 0;

    my @seq1;
    my @seq2;
    
    my @numl1;
    my @numl2;

    my @numr1;
    my @numr2;

    open (OUT, ">$fileq") || die;

    open (SBLAST, "$file") || die;

    while (<SBLAST>) {
	if    (/^>(\S+)/ && $seq == 0) { 
	    if ($nseq1 > 0) {
		if ($numl1[0] < $numr1[$line1-1]) { $start1 = $numl1[0].">".$numr1[$line1-1]; }
		else                              { $start1 = $numr1[$line1-1]."<".$numl1[0]; }
		
		print OUT ">", $name1, "-", $start1, "-", $annotation, "\n";
		for (my $i = 0; $i < $line1; $i++) {
		    print OUT "\t", $seq1[$i], "\n"; }
	    }
	    $line1 = 0;
	    $name1 = $1;
	    $seq = 1;
	    $nseq1++;
	}
	
	elsif (/^>(\S+)/ && $seq == 1) {
	    if ($nseq2 > 0) {
		if ($numl2[0] < $numr2[$line2-1]) { $start2 = $numl2[0].">".$numr2[$line2-1]; }
		else                              { $start2 = $numr2[$line2-1]."<".$numl2[0]; }
		
		print OUT ">", $name2, "-", $start2, "-", $annotation, "\n";
		for (my $i = 0; $i < $line2; $i++) {
		    print OUT "\t", $seq2[$i], "\n"; }
	    }
	    $line2 = 0;
	    $name2 = $1;
	    $seq = 0;
	    $nseq2++;
	    $annotation = ""; 
	}
	elsif (/^Query:\s+(\S+)\s+(\S+)\s(\S+)/) { 
	    $numl1[$line1] = $1; $seq1[$line1] = $2; $numr1[$line1] = $3; $line1++; }
	elsif (/^Sbjct:\s+(\S+)\s+(\S+)\s(\S+)/) { 
	    $numl2[$line2] = $1; $seq2[$line2] = $2; $numr2[$line2] = $3; $line2++; }
	elsif (/^\s+Annotation =\s(.+)/  ) { 
	    $more = $1;  @more_parts = split(/ /, $more);
	    $info = join("**", @more_parts);
	    $annotation = $annotation.$info; }
	else { next; }
    }
    
    close (SBLAST);
}

# depile_blastn ()
#
#
sub depile_blastn {

    my ($fileb, $filec) = @_;

    my $keyquery;
    my $keyquery_new;

    my $numl;
    my $numr;

    my $start;
    my $end;

    my @howmany;
    my @which;

    my @array;
    my $pile;

    my %bit;

    my $count_query = 0;
    my $count_ali_per_query = 0;

    my $queryplot;

    my $num;

    open (SBLAST, "$fileb") || die;
    while (<SBLAST>) {
	if    (/^\s*Query=\s+(\S+)/) 
	{ 
	    $keyquery_new = $1; 

	    if ($count_query > 0) {

		for (my $n = 0; $n < $n_ali_per_query_one{$keyquery}; $n++) { $bit{$keyquery} .= "0"; }

		if ($verbose) { 
		    $keyquery =~ /^(\d+)>(\S+)$/;
		    $queryplot = "$1_$2.plot"; 
		    system ("touch $queryplot\n"); 
		    open (PLOT, ">$queryplot") || die; 
		}

		my $pos = 0;
		while ($pos < $len_query{$keyquery}) {
			
		    $pile = ($howmany[$pos] > $depth)? $depth : $howmany[$pos];

		    if ($verbose) { print PLOT "$pos $howmany[$pos] $pile\n"; }

		    @array = split(/-/, $which[$pos]);
		    
		    #paranoia
		    if (($#array+1) != $howmany[$pos]) { print "how many alignments at position $pos? ", $#array+1, " or $howmany[$pos]?\n"; die; }
		    
		    for (my $num = 0; $num < $pile; $num++) {
			substr($bit{$keyquery}, $array[$num], 1) = "1";
		    }

		    $pos += $shift;
		}

		if ($verbose) { close (PLOT); plot ("$queryplot"); }
 	    }

	    $count_query ++; 
	    $count_ali_per_query = 0;
	    $keyquery = $keyquery_new;

	    for (my $l = 0; $l < $len_query{$keyquery}; $l++) { $howmany[$l] = 0; $which[$l] = "";}
	    
	}    
	elsif (/^\s+Identities/)       
	{ 
	    $count_ali_per_query ++;
	}
	elsif (/^Query:\s+(\S+)\s\S+\s(\S+)/) 
	{ 
	    
	    $numl = $1; 
	    $numr = $2;
	    
	    if ($numl < $numr) { $start = $numl-1; $end = $numr-1; }
	    else               { $start = $numr-1; $end = $numl-1; }
	    
	    if ($start < 0 || $end >= $len_query{$keyquery}) { die "wrong blast limits\n"; }

	    for (my $l = $start; $l <= $end; $l++) {
		$howmany[$l] ++;
		$num = $count_ali_per_query - 1;
		$which[$l] .= $num."-";
	    }
	}
    }
    
    close (SBLAST);

    for (my $n = 0; $n < $n_ali_per_query_one{$keyquery}; $n++) { $bit{$keyquery} .= "0"; }

    if ($verbose) { 
	$keyquery =~ /^(\d+)>(\S+)$/;
	$queryplot = "$1_$2.plot"; 
 	system ("touch $queryplot\n"); 
	open (PLOT, ">$queryplot") || die; 
    }

    my $pos = 0;
    while ($pos < $len_query{$keyquery}) {

	$pile = ($howmany[$pos] > $depth)? $depth : $howmany[$pos];
	
	if ($verbose) { print PLOT "$pos $howmany[$pos] $pile\n"; }
	
	@array = split(/-/, $which[$pos]);
	
	#paranoia
	if (($#array+1) != $howmany[$pos]) { print "how many alignments at position $pos? ", $#array+1, " or $howmany[$pos]?\n"; die; }
	
	for (my $num = 0; $num < $pile; $num++) {
	    substr($bit{$keyquery}, $array[$num], 1) = "1";
	}
	
	$pos += $shift;
    }

    if ($verbose) { close (PLOT); plot ("$queryplot"); }
    
    
    
    # now, print the selected alignments
    
    my $flag = 0;
    my $count = 0;
    my $count_two = 0;

    my @qbit;
    my $query;
    my $sbjct;

    my $covered;
    my $uncovered;
    my $total;

    open (OUT, ">$filec") || die;
    
    open (SBLAST, "$fileb") || die;
    while (<SBLAST>) {
	if    (/^\s*Query=\s+(\S+)/) 
	{ 
	    $keyquery = $1; 

	    $keyquery =~ /^\d+>(\S+)$/; $query = $1;

	    $covered   = ($bit{$keyquery} =~ s/1/1/g);
	    $uncovered = ($bit{$keyquery} =~ s/0/0/g);
	    $total = $covered + $uncovered;  
    
	    $n_ali_per_query_two{$keyquery}  = $covered;
	    $n_ali_two                      += $covered;


	    $count     = 0;
	    $count_two = 0;

	    @qbit = split(//,$bit{$keyquery});

	    #paranoia
	    if ($total   != $n_ali_per_query_one{$keyquery}) { print "how many at first at $keyquery? $total or $n_ali_per_query_one{$keyquery}?\n";  die; }
	    if ($covered != $n_ali_per_query_two{$keyquery}) { print "how many at last at $keyquery? $covered or $n_ali_per_query_two{$keyquery}?\n";  die; }
	    if ($#qbit+1 != $n_ali_per_query_one{$keyquery}) { print "how many in qbit at $keyquery? ", $#qbit+1, " or $n_ali_per_query_one{$keyquery}?\n"; die; }

	}
	elsif (/^>(\S+)/)       
	{ 
	    $sbjct = $1;
	    $flag = $qbit[$count];
	    $count ++;
	    
	    if ($flag) {
		$count_two ++;

		print OUT ">$query\n";
		print OUT ">$sbjct\n";
	    }

	}
	elsif ($flag && /^\s+Identities = \S+\/(\S+)/) 
	{ 
	    $ave_len_ali_two += $1;
	    print OUT $_; 
	}
	elsif ($flag) { print OUT $_; }

    }
    
    close (SBLAST);

    #paranoia
    if ($count     != $n_ali_per_query_one{$keyquery}) { print "how many at first for sure at $keyquery? $count or $n_ali_per_query_one{$keyquery}?\n"; die; }
    if ($count_two != $n_ali_per_query_two{$keyquery}) { print "how many at last for sure at $keyquery? $count_two or $n_ali_per_query_two{$keyquery}?\n"; die; }
}



# order_blastn ()
#
#
sub order_blastn {

    my ($filea, $fileb) = @_;

    my $keyquery;
    my $keyquery_new;

    my $sbjct;

    my $count_qr= 0;

    my $n_ali= 0;
    my @ali;
    my @sc;

    my $sc;
    my $ali;

    open (OUT, ">$fileb") || die;

    open (SBLAST, "$filea") || die;
    while (<SBLAST>) {
	if    (/^\s*Query=\s+(\S+)/) 
	{ 
	    $keyquery_new = $1; 

	    if ($count_qr > 0) {
		for (my $i = 0; $i <= $n_ali; $i++) {
		    for (my $j = $i+1; $j <= $n_ali; $j++) {
			if ($sc[$i] < $sc[$j]) {
			    $sc  = $sc[$i];
			    $ali = $ali[$i];

			    $sc[$i]  = $sc[$j];
			    $ali[$i] = $ali[$j];

			    $sc[$j]  = $sc;
			    $ali[$j] = $ali;

			}
		    }
		}

		print OUT "Query=  $keyquery\n";
		for (my $i = 0; $i <= $n_ali; $i++) {
		    print OUT $ali[$i] ;
		}
	    }
	    
	    $n_ali = -1;
	    undef @ali;
	    $keyquery = $keyquery_new;
	    $count_qr ++;
	}	
	elsif (/^>(\S+)/) 
	{
	    $n_ali ++;
	    $ali[$n_ali] .= $_; 
	}
	elsif ($which =~ /^ID$/ && /^\s+Identities = \S+\/\S+\s\((\S+)%\)/) 
	{
     	    $sc[$n_ali] = $1; 
	    $ali[$n_ali] .= $_; 
	}
	elsif ($which =~ /^SC$/ && /^\s+Score = (\S+)/) 
	{
     	    $sc[$n_ali] = $1; 
	    $ali[$n_ali] .= $_; 
	}
	elsif ($n_ali > -1) { $ali[$n_ali] .= $_; }
    }
    
    for (my $i = 0; $i <= $n_ali; $i++) {
	for (my $j = $i+1; $j <= $n_ali; $j++) {
	    if ($sc[$i] < $sc[$j]) {
		$sc  = $sc[$i];
		$ali = $ali[$i];
		
		$sc[$i]  = $sc[$j];
		$ali[$i] = $ali[$j];
		
		$sc[$j]  = $sc;
		$ali[$j] = $ali;
		
	    }
	}
    }
    
    print OUT "Query=  $keyquery\n";
    for (my $i = 0; $i <= $n_ali; $i++) {
	print OUT $ali[$i] ;
    }
    
    close (SBLAST);
    close (OUT);

    #paranoia
    if ($n_ali+1 != $n_ali_per_query_one{$keyquery}) { print "how many after ordering? ", $n_ali+1, " or $n_ali_per_query_one{$keyquery}?\n"; die; }
}


# plot ()
#
#
sub plot {

    my ($plotfile) = @_;

    open(GP,'|'.GNUPLOT) || die "Gnuplot: $!";
    
    print GP "set terminal postscript color\n";
    print GP "set output '$plotfile.ps'\n";
    
    
    print GP "set title 'Density of BLASTN alignments'\n"; 
    print GP "set xlabel 'Position'\n";
    print GP "set ylabel 'Number of blastn hits'\n";
    
    print GP "plot '$plotfile' using 1:2 with lines,  '$plotfile' using 1:3 with lines\n";
    close GP;
    
    system ("ghostview -landscape -magstep -2 $plotfile.ps&\n");
}

# prune_blastn ()
#
#
sub prune_blastn {

    my ($file, $filea) = @_;

    my $query;
    my $keyquery;
    my $sbjct;

    my $length;

    my $evalue;
    my $scoreline;

    my $identity;
    my $identityline;

    my $flag = 0;

    open (OUT, ">$filea") || die;

    open (BLAST, "$file") || die;
    while (<BLAST>) {

	if    (/^\s*Query=\s+(\S+)/) 
	{ 
	    $n_query ++;
	    $query = $1; 
	    $keyquery = $n_query.">".$query; 

	    print OUT "Query=  $keyquery\n";

	    #set counters
	    $n_ali_per_query_total{$keyquery} = 0;
	    $n_ali_per_query_one{$keyquery}   = 0;
	    $n_ali_per_query_two{$keyquery}   = 0;
	}
	elsif (/^\s+\((\S+) letters/) 
	{ 
	    $len_query{$keyquery} = $1;  $len_query{$keyquery} =~ s/,//g;
	}
	elsif (/^>(\S+)/)       
	{ 
	    $sbjct = $1; $sbjct =~ s/[\(\)\|\+\_\/]/-/g;  
	}
	elsif (/Expect = (\S+),/) 
	{
	    $evalue = $1;
	    $scoreline = $_; 
	}
	elsif (/^\s+Identities = \S+\/(\S+)\s\((\S+)%\)/) 
	{
	    $flag = 0;

	    $length   = $1; 
	    $identity = $2; 
	    
	    $identityline = $_; 

	    $n_ali_total  ++;
	    $n_ali_per_query_total{$keyquery}  ++;

	    $ave_len_ali_total += $length;

	    if ($length   >= $min_len && $length   <= $max_len &&
		$identity >= $min_id  && $identity <= $max_id  &&  
		$evalue   <= $max_eval) { 
		$n_ali_per_query_one{$keyquery} ++; $n_ali_one ++; 
		$flag = 1;

		$ave_len_ali_one += $length;
	
		print OUT ">", $sbjct, "\n";
		print OUT $scoreline;
		print OUT $identityline;
	    }
	
	}
	elsif ($flag) { print OUT $_; }
	
    }
    close (BLAST);
    close (OUT);
}


# write_report ()
#
#
sub write_report {
    my ($report) = @_;

    my $query;
    my $num;

    open (REP, ">$report") || die;
    
    print REP "FILE: \t$file\n";
    print REP "DIR:  \t$dir/\n";


    print REP "\nFIRST TRIMMING\n";
    print REP "Minimum length = $min_len\n";
    print REP "Maximum Evalue = $max_eval\n";
    print REP "Minimum %id    = $min_id\n";
    print REP "Maximum %id    = $max_id\n";

    print REP "\nSECOND TRIMMING\n";
    print REP "Alignments culled by = $which\n";
    print REP "Depth of alignments  = $depth\n";
    print REP "shift                = $shift\n\n";

    foreach my $r (keys(%n_ali_per_query_total)) { 

	$r =~ /^(\d+)>(\S+)$/; $num = $1; $query = $2; 
	print REP "$num-QUERY: $query \n";
	print REP "\tTotal \# alignments: $n_ali_per_query_total{$r}\t After First trimming: $n_ali_per_query_one{$r}\t After Second trimming: $n_ali_per_query_two{$r} \n";
    }

    printf (REP "\nTotal \#Queries     \t%d\n", $n_query);
    printf (REP "Total \#Alignments    \t%d\t", $n_ali_total);
    if ($n_ali_total > 0.0) { printf (REP " ave_len = %.1f\n", $ave_len_ali_total/$n_ali_total); }
    else { printf (REP " ave_len = 0\n"); }
    
    printf (REP "After first trimming  \t%d\t", $n_ali_one);
    if ($n_ali_one > 0.0) { printf (REP " ave_len = %.1f\n", $ave_len_ali_one/$n_ali_one); }
    else { printf (REP " ave_len = 0\n"); }
    
    printf (REP "After second trimming \t%d\t", $n_ali_two);
    if ($n_ali_two > 0.0) {  printf (REP " ave_len = %.1f\n", $ave_len_ali_two/$n_ali_two); }
    else { printf (REP " ave_len = 0\n"); }
   
    close (REP);
}



